@@ -11,6 +11,7 @@ import (
 
 	"github.com/go-logr/logr"
 	"golang.org/x/sync/semaphore"
+	v1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
@@ -24,17 +25,22 @@ import (
 	"k8s.io/client-go/tools/pager"
 	watchutil "k8s.io/client-go/tools/watch"
 	"k8s.io/klog/v2/klogr"
+	"k8s.io/kubectl/pkg/util/openapi"
 
 	"github.com/argoproj/gitops-engine/pkg/utils/kube"
 	"github.com/argoproj/gitops-engine/pkg/utils/tracing"
 )
 
 const (
-	clusterResyncTimeout       = 24 * time.Hour
-	watchResyncTimeout         = 10 * time.Minute
 	watchResourcesRetryTimeout = 1 * time.Second
 	ClusterRetryTimeout        = 10 * time.Second
 
+	// default duration before we invalidate entire cluster cache. Can be set to 0 to never invalidate cache
+	defaultClusterResyncTimeout = 24 * time.Hour
+
+	// default duration before restarting individual resource watch
+	defaultWatchResyncTimeout = 10 * time.Minute
+
 	// Same page size as in k8s.io/client-go/tools/pager/pager.go
 	defaultListPageSize = 500
 	// Prefetch only a single page
@@ -65,6 +71,8 @@ type ClusterInfo struct {
 	LastCacheSyncTime *time.Time
 	// SyncError holds most recent cache synchronization error
 	SyncError error
+	// APIResources holds list of API resources supported by the cluster
+	APIResources []kube.APIResourceInfo
 }
 
 // OnEventHandler is a function that handles Kubernetes event
@@ -82,12 +90,14 @@ type ClusterCache interface {
 	EnsureSynced() error
 	// GetServerVersion returns observed cluster version
 	GetServerVersion() string
-	// GetAPIGroups returns information about observed API groups
-	GetAPIGroups() []metav1.APIGroup
+	// GetAPIResources returns information about observed API resources
+	GetAPIResources() []kube.APIResourceInfo
+	// GetOpenAPISchema returns open API schema of supported API resources
+	GetOpenAPISchema() openapi.Resources
 	// Invalidate cache and executes callback that optionally might update cache settings
 	Invalidate(opts ...UpdateSettingsFunc)
-	// GetNamespaceTopLevelResources returns top level resources in the specified namespace
-	GetNamespaceTopLevelResources(namespace string) map[kube.ResourceKey]*Resource
+	// FindResources returns resources that matches given list of predicates from specified namespace or everywhere if specified namespace is empty
+	FindResources(namespace string, predicates ...func(r *Resource) bool) map[kube.ResourceKey]*Resource
 	// IterateHierarchy iterates resource tree starting from the specified top level resource and executes callback for each resource in the tree
 	IterateHierarchy(key kube.ResourceKey, action func(resource *Resource, namespaceResources map[kube.ResourceKey]*Resource))
 	// IsNamespaced answers if specified group/kind is a namespaced resource API or not
@@ -114,7 +124,6 @@ type WeightedSemaphore interface {
 func NewClusterCache(config *rest.Config, opts ...UpdateSettingsFunc) *clusterCache {
 	log := klogr.New()
 	cache := &clusterCache{
-		resyncTimeout:      clusterResyncTimeout,
 		settings:           Settings{ResourceHealthOverride: &noopSettings{}, ResourcesFilter: &noopSettings{}},
 		apisMeta:           make(map[schema.GroupKind]*apiMeta),
 		listPageSize:       defaultListPageSize,
@@ -127,7 +136,11 @@ func NewClusterCache(config *rest.Config, opts ...UpdateSettingsFunc) *clusterCa
 			Log:    log,
 			Tracer: tracing.NopTracer{},
 		},
-		syncTime:                nil,
+		syncStatus: clusterCacheSync{
+			resyncTimeout: defaultClusterResyncTimeout,
+			syncTime:      nil,
+		},
+		watchResyncTimeout:      defaultWatchResyncTimeout,
 		resourceUpdatedHandlers: map[uint64]OnResourceUpdatedHandler{},
 		eventHandlers:           map[uint64]OnEventHandler{},
 		log:                     log,
@@ -139,15 +152,17 @@ func NewClusterCache(config *rest.Config, opts ...UpdateSettingsFunc) *clusterCa
 }
 
 type clusterCache struct {
-	resyncTimeout time.Duration
-	syncTime      *time.Time
-	syncError     error
+	syncStatus clusterCacheSync
+
 	apisMeta      map[schema.GroupKind]*apiMeta
 	serverVersion string
-	apiGroups     []metav1.APIGroup
+	apiResources  []kube.APIResourceInfo
 	// namespacedResources is a simple map which indicates a groupKind is namespaced
 	namespacedResources map[schema.GroupKind]bool
 
+	// maximum time we allow watches to run before relisting the group/kind and restarting the watch
+	watchResyncTimeout time.Duration
+
 	// size of a page for list operations pager.
 	listPageSize int64
 	// number of pages to prefetch for list pager.
@@ -159,17 +174,30 @@ type clusterCache struct {
 	resources map[kube.ResourceKey]*Resource
 	nsIndex   map[string]map[kube.ResourceKey]*Resource
 
-	kubectl    kube.Kubectl
-	log        logr.Logger
-	config     *rest.Config
-	namespaces []string
-	settings   Settings
+	kubectl          kube.Kubectl
+	log              logr.Logger
+	config           *rest.Config
+	namespaces       []string
+	clusterResources bool
+	settings         Settings
 
 	handlersLock                sync.Mutex
 	handlerKey                  uint64
 	populateResourceInfoHandler OnPopulateResourceInfoHandler
 	resourceUpdatedHandlers     map[uint64]OnResourceUpdatedHandler
 	eventHandlers               map[uint64]OnEventHandler
+	openAPISchema               openapi.Resources
+}
+
+type clusterCacheSync struct {
+	// When using this struct:
+	// 1) 'lock' mutex should be acquired when reading/writing from fields of this struct.
+	// 2) The parent 'clusterCache.lock' does NOT need to be owned to r/w from fields of this struct (if it is owned, that is fine, but see below)
+	// 3) To prevent deadlocks, do not acquire parent 'clusterCache.lock' after acquiring this lock; if you need both locks, always acquire the parent lock first
+	lock          sync.Mutex
+	syncTime      *time.Time
+	syncError     error
+	resyncTimeout time.Duration
 }
 
 // OnResourceUpdated register event handler that is executed every time when resource get's updated in the cache
@@ -225,9 +253,40 @@ func (c *clusterCache) GetServerVersion() string {
 	return c.serverVersion
 }
 
-// GetAPIGroups returns information about observed API groups
-func (c *clusterCache) GetAPIGroups() []metav1.APIGroup {
-	return c.apiGroups
+// GetAPIResources returns information about observed API resources
+func (c *clusterCache) GetAPIResources() []kube.APIResourceInfo {
+	c.lock.RLock()
+	defer c.lock.RUnlock()
+
+	return c.apiResources
+}
+
+// GetOpenAPISchema returns open API schema of supported API resources
+func (c *clusterCache) GetOpenAPISchema() openapi.Resources {
+	return c.openAPISchema
+}
+
+func (c *clusterCache) appendAPIResource(info kube.APIResourceInfo) {
+	exists := false
+	for i := range c.apiResources {
+		if c.apiResources[i].GroupKind == info.GroupKind && c.apiResources[i].GroupVersionResource.Version == info.GroupVersionResource.Version {
+			exists = true
+			break
+		}
+	}
+	if !exists {
+		c.apiResources = append(c.apiResources, info)
+	}
+}
+
+func (c *clusterCache) deleteAPIResource(info kube.APIResourceInfo) {
+	for i := range c.apiResources {
+		if c.apiResources[i].GroupKind == info.GroupKind && c.apiResources[i].GroupVersionResource.Version == info.GroupVersionResource.Version {
+			c.apiResources[i] = c.apiResources[len(c.apiResources)-1]
+			c.apiResources = c.apiResources[:len(c.apiResources)-1]
+			break
+		}
+	}
 }
 
 func (c *clusterCache) replaceResourceCache(gk schema.GroupKind, resources []*Resource, ns string) {
@@ -312,7 +371,11 @@ func (c *clusterCache) setNode(n *Resource) {
 func (c *clusterCache) Invalidate(opts ...UpdateSettingsFunc) {
 	c.lock.Lock()
 	defer c.lock.Unlock()
-	c.syncTime = nil
+
+	c.syncStatus.lock.Lock()
+	c.syncStatus.syncTime = nil
+	c.syncStatus.lock.Unlock()
+
 	for i := range c.apisMeta {
 		c.apisMeta[i].watchCancel()
 	}
@@ -324,15 +387,21 @@ func (c *clusterCache) Invalidate(opts ...UpdateSettingsFunc) {
 	c.log.Info("Invalidated cluster")
 }
 
-func (c *clusterCache) synced() bool {
-	syncTime := c.syncTime
+// clusterCacheSync's lock should be held before calling this method
+func (syncStatus *clusterCacheSync) synced() bool {
+	syncTime := syncStatus.syncTime
+
 	if syncTime == nil {
 		return false
 	}
-	if c.syncError != nil {
+	if syncStatus.syncError != nil {
 		return time.Now().Before(syncTime.Add(ClusterRetryTimeout))
 	}
-	return time.Now().Before(syncTime.Add(c.resyncTimeout))
+	if syncStatus.resyncTimeout == 0 {
+		// cluster resync timeout has been disabled
+		return true
+	}
+	return time.Now().Before(syncTime.Add(syncStatus.resyncTimeout))
 }
 
 func (c *clusterCache) stopWatching(gk schema.GroupKind, ns string) {
@@ -348,7 +417,7 @@ func (c *clusterCache) stopWatching(gk schema.GroupKind, ns string) {
 
 // startMissingWatches lists supported cluster resources and start watching for changes unless watch is already running
 func (c *clusterCache) startMissingWatches() error {
-	apis, err := c.kubectl.GetAPIResources(c.config, c.settings.ResourcesFilter)
+	apis, err := c.kubectl.GetAPIResources(c.config, true, c.settings.ResourcesFilter)
 	if err != nil {
 		return err
 	}
@@ -457,8 +526,12 @@ func (c *clusterCache) watchEvents(ctx context.Context, api kube.APIResourceInfo
 			resourceVersion = ""
 		}()
 
-		shouldResync := time.NewTimer(watchResyncTimeout)
-		defer shouldResync.Stop()
+		var watchResyncTimeoutCh <-chan time.Time
+		if c.watchResyncTimeout > 0 {
+			shouldResync := time.NewTimer(c.watchResyncTimeout)
+			defer shouldResync.Stop()
+			watchResyncTimeoutCh = shouldResync.C
+		}
 
 		for {
 			select {
@@ -467,7 +540,7 @@ func (c *clusterCache) watchEvents(ctx context.Context, api kube.APIResourceInfo
 				return nil
 
 			// re-synchronize API state and restart watch periodically
-			case <-shouldResync.C:
+			case <-watchResyncTimeoutCh:
 				return fmt.Errorf("Resyncing %s on %s during to timeout", api.GroupKind, c.config.Host)
 
 			// re-synchronize API state and restart watch if retry watcher failed to continue watching using provided resource version
@@ -486,15 +559,41 @@ func (c *clusterCache) watchEvents(ctx context.Context, api kube.APIResourceInfo
 
 				c.processEvent(event.Type, obj)
 				if kube.IsCRD(obj) {
-					if event.Type == watch.Deleted {
-						group, groupOk, groupErr := unstructured.NestedString(obj.Object, "spec", "group")
-						kind, kindOk, kindErr := unstructured.NestedString(obj.Object, "spec", "names", "kind")
+					var resources []kube.APIResourceInfo
+					crd := v1.CustomResourceDefinition{}
+					err := runtime.DefaultUnstructuredConverter.FromUnstructured(obj.Object, &crd)
+					if err != nil {
+						c.log.Error(err, "Failed to extract CRD resources")
+					}
+					for _, v := range crd.Spec.Versions {
+						resources = append(resources, kube.APIResourceInfo{
+							GroupKind: schema.GroupKind{
+								Group: crd.Spec.Group, Kind: crd.Spec.Names.Kind},
+							GroupVersionResource: schema.GroupVersionResource{
+								Group: crd.Spec.Group, Version: v.Name, Resource: crd.Spec.Names.Plural},
+							Meta: metav1.APIResource{
+								Group:        crd.Spec.Group,
+								SingularName: crd.Spec.Names.Singular,
+								Namespaced:   crd.Spec.Scope == v1.NamespaceScoped,
+								Name:         crd.Spec.Names.Plural,
+								Kind:         crd.Spec.Names.Singular,
+								Version:      v.Name,
+								ShortNames:   crd.Spec.Names.ShortNames,
+							},
+						})
+					}
 
-						if groupOk && groupErr == nil && kindOk && kindErr == nil {
-							gk := schema.GroupKind{Group: group, Kind: kind}
-							c.stopWatching(gk, ns)
+					if event.Type == watch.Deleted {
+						for i := range resources {
+							c.deleteAPIResource(resources[i])
 						}
 					} else {
+						// add new CRD's groupkind to c.apigroups
+						if event.Type == watch.Added {
+							for i := range resources {
+								c.appendAPIResource(resources[i])
+							}
+						}
 						err = runSynced(&c.lock, func() error {
 							return c.startMissingWatches()
 						})
@@ -502,6 +601,17 @@ func (c *clusterCache) watchEvents(ctx context.Context, api kube.APIResourceInfo
 							c.log.Error(err, "Failed to start missing watch")
 						}
 					}
+					err = runSynced(&c.lock, func() error {
+						openAPISchema, err := c.kubectl.LoadOpenAPISchema(c.config)
+						if err != nil {
+							return err
+						}
+						c.openAPISchema = openAPISchema
+						return nil
+					})
+					if err != nil {
+						c.log.Error(err, "Failed to reload open api schema")
+					}
 				}
 			}
 		}
@@ -510,20 +620,20 @@ func (c *clusterCache) watchEvents(ctx context.Context, api kube.APIResourceInfo
 
 func (c *clusterCache) processApi(client dynamic.Interface, api kube.APIResourceInfo, callback func(resClient dynamic.ResourceInterface, ns string) error) error {
 	resClient := client.Resource(api.GroupVersionResource)
-	if len(c.namespaces) == 0 {
+	switch {
+	// if manage whole cluster or resource is cluster level and cluster resources enabled
+	case len(c.namespaces) == 0 || !api.Meta.Namespaced && c.clusterResources:
 		return callback(resClient, "")
-	}
-
-	if !api.Meta.Namespaced {
-		return nil
-	}
-
-	for _, ns := range c.namespaces {
-		err := callback(resClient.Namespace(ns), ns)
-		if err != nil {
-			return err
+	// if manage some namespaces and resource is namespaced
+	case len(c.namespaces) != 0 && api.Meta.Namespaced:
+		for _, ns := range c.namespaces {
+			err := callback(resClient.Namespace(ns), ns)
+			if err != nil {
+				return err
+			}
 		}
 	}
+
 	return nil
 }
 
@@ -543,12 +653,19 @@ func (c *clusterCache) sync() error {
 		return err
 	}
 	c.serverVersion = version
-	groups, err := c.kubectl.GetAPIGroups(config)
+	apiResources, err := c.kubectl.GetAPIResources(config, false, NewNoopSettings())
 	if err != nil {
 		return err
 	}
-	c.apiGroups = groups
-	apis, err := c.kubectl.GetAPIResources(c.config, c.settings.ResourcesFilter)
+	c.apiResources = apiResources
+
+	openAPISchema, err := c.kubectl.LoadOpenAPISchema(config)
+	if err != nil {
+		return err
+	}
+	c.openAPISchema = openAPISchema
+
+	apis, err := c.kubectl.GetAPIResources(c.config, true, c.settings.ResourcesFilter)
 
 	if err != nil {
 		return err
@@ -601,36 +718,62 @@ func (c *clusterCache) sync() error {
 
 // EnsureSynced checks cache state and synchronizes it if necessary
 func (c *clusterCache) EnsureSynced() error {
-	// first check if cluster is synced *without lock*
-	if c.synced() {
-		return c.syncError
+	syncStatus := &c.syncStatus
+
+	// first check if cluster is synced *without acquiring the full clusterCache lock*
+	syncStatus.lock.Lock()
+	if syncStatus.synced() {
+		syncError := syncStatus.syncError
+		syncStatus.lock.Unlock()
+		return syncError
 	}
+	syncStatus.lock.Unlock() // release the lock, so that we can acquire the parent lock (see struct comment re: lock acquisition ordering)
 
 	c.lock.Lock()
 	defer c.lock.Unlock()
+	syncStatus.lock.Lock()
+	defer syncStatus.lock.Unlock()
+
 	// before doing any work, check once again now that we have the lock, to see if it got
 	// synced between the first check and now
-	if c.synced() {
-		return c.syncError
+	if syncStatus.synced() {
+		return syncStatus.syncError
 	}
 	err := c.sync()
 	syncTime := time.Now()
-	c.syncTime = &syncTime
-	c.syncError = err
-	return c.syncError
+	syncStatus.syncTime = &syncTime
+	syncStatus.syncError = err
+	return syncStatus.syncError
 }
 
-// GetNamespaceTopLevelResources returns top level resources in the specified namespace
-func (c *clusterCache) GetNamespaceTopLevelResources(namespace string) map[kube.ResourceKey]*Resource {
+func (c *clusterCache) FindResources(namespace string, predicates ...func(r *Resource) bool) map[kube.ResourceKey]*Resource {
 	c.lock.RLock()
 	defer c.lock.RUnlock()
-	resources := make(map[kube.ResourceKey]*Resource)
-	for _, res := range c.nsIndex[namespace] {
-		if len(res.OwnerRefs) == 0 {
-			resources[res.ResourceKey()] = res
+	result := map[kube.ResourceKey]*Resource{}
+	resources := map[kube.ResourceKey]*Resource{}
+	if namespace != "" {
+		if ns, ok := c.nsIndex[namespace]; ok {
+			resources = ns
+		}
+	} else {
+		resources = c.resources
+	}
+
+	for k := range resources {
+		r := resources[k]
+		matches := true
+		for i := range predicates {
+			if !predicates[i](r) {
+				matches = false
+				break
+			}
+		}
+
+		if matches {
+			result[k] = r
 		}
 	}
-	return resources
+	return result
 }
 
 // IterateHierarchy iterates resource tree starting from the specified top level resource and executes callback for each resource in the tree
@@ -679,12 +822,6 @@ func (c *clusterCache) IsNamespaced(gk schema.GroupKind) (bool, error) {
 }
 
 func (c *clusterCache) managesNamespace(namespace string) bool {
-	if len(c.namespaces) == 0 {
-		return true
-	}
-	if namespace == "" {
-		return false
-	}
 	for _, ns := range c.namespaces {
 		if ns == namespace {
 			return true
@@ -701,11 +838,12 @@ func (c *clusterCache) GetManagedLiveObjs(targetObjs []*unstructured.Unstructure
 	defer c.lock.RUnlock()
 
 	for _, o := range targetObjs {
-		if !c.managesNamespace(o.GetNamespace()) {
-			if o.GetNamespace() == "" {
+		if len(c.namespaces) > 0 {
+			if o.GetNamespace() == "" && !c.clusterResources {
 				return nil, fmt.Errorf("Cluster level %s %q can not be managed when in namespaced mode", o.GetKind(), o.GetName())
+			} else if o.GetNamespace() != "" && !c.managesNamespace(o.GetNamespace()) {
+				return nil, fmt.Errorf("Namespace %q for %s %q is not managed", o.GetNamespace(), o.GetKind(), o.GetName())
 			}
-			return nil, fmt.Errorf("Namespace %q for %s %q is not managed", o.GetNamespace(), o.GetKind(), o.GetName())
 		}
 	}
 
@@ -842,13 +980,17 @@ var (
 func (c *clusterCache) GetClusterInfo() ClusterInfo {
 	c.lock.RLock()
 	defer c.lock.RUnlock()
+	c.syncStatus.lock.Lock()
+	defer c.syncStatus.lock.Unlock()
+
 	return ClusterInfo{
 		APIsCount:         len(c.apisMeta),
 		K8SVersion:        c.serverVersion,
 		ResourcesCount:    len(c.resources),
 		Server:            c.config.Host,
-		LastCacheSyncTime: c.syncTime,
-		SyncError:         c.syncError,
+		LastCacheSyncTime: c.syncStatus.syncTime,
+		SyncError:         c.syncStatus.syncError,
+		APIResources:      c.apiResources,
 	}
 }
 
